function rm = GPregressionM(input,obs,obs1,kernel,options,basisFunc,dBasisFunc)
%GPregression Computes GP regression model, contains a function for
%prediction and a function 

rm.input = input;
rm.kernel = kernel;
rm.basisFunc = basisFunc;
rm.dBasisFunc = dBasisFunc;

rm.dim = size(input,2);
rm.n = size(input,1);
rm.q = length(basisFunc(input(1,:)));

%d is the alpha-parameter of the inverse-normal gamma distribution, and
%should be set ot zero to get the uninformative prior with infinite variance 
%on sigma2.
rm.d = 0;

if nargin <= 5
   rm.basisFunc = @(X)[1]; 
   rm.dBasisFunc = @(X)[0]; 
end


%check data consistency
if (size(obs,2) ~= rm.n)
    disp('input data and observations have different dimesions');
    return;
end
if (kernel.dim ~= rm.dim)
    disp('kernel has the wrong dimensions');
    return;
end

%Process input, computing the necessary matrices
%compute gram matrix and H matrix

if kernel.derivative
    rm.N = rm.n*(1+rm.dim);
    rm.step = 1+rm.dim;
else
    rm.N = rm.n;
    rm.step = 1;
end
rm.K = zeros(rm.N);
rm.H = zeros(rm.N,rm.q);
for i = 1:rm.n
    I = 1 + (i-1)*rm.step;
    if kernel.derivative
       rm.H(I:(I+rm.step-1),:) = [basisFunc(input(i,:));
                                dBasisFunc(input(i,:))]; 
    else
       rm.H(I,:) = basisFunc(input(i,:));
    end
    for j = 1:rm.n
        J = 1 + (j-1)*rm.step; 
        rm.K(I:(I+rm.step-1),J:(J+rm.step-1)) = kernel.evalGram(input(i,:),input(j,:),kernel);
    end
end
%forcing the matrix to be positive definite by adding a small noise term
rm.K = rm.K + eye(rm.N)*options.noise_level;

rm.L = chol(rm.K,'lower');
rm.Kinv = inv(rm.L')*inv(rm.L);
%KinvH = L'\(L\H);
rm.KinvH = rm.Kinv*rm.H;
V_star_inv = rm.H'*rm.KinvH; 
Lstar = chol(V_star_inv,'lower');
rm.V_star = inv(Lstar')*inv(Lstar);

rm.conf = tinv((1+options.confidence_level)/2,rm.n+rm.d);

if kernel.derivative
    rm.y = zeros(rm.N,1);
    for i=1:rm.n
        I = 1 + (i-1)*rm.step;
        rm.y(I:(I+rm.step-1)) = [obs(i);obs1(1,:,i)'];
    end
else
    rm.y = obs';
end
%Kobs = L'\(L\obs);
rm.Kobs = rm.Kinv*rm.y;
%beta_hat = Lstar'\(Lstar\(H'*Kobs));
rm.beta_hat = rm.V_star*(rm.H'*rm.Kobs);
sigma2_hat = (rm.y'*rm.Kobs - rm.beta_hat'*V_star_inv*rm.beta_hat)/(rm.n+rm.d-2);
if sigma2_hat < 0
    sigma2_hat = 1e-16;
end
rm.sigma_hat = sqrt(sigma2_hat);
%rm.sigma_hat = sqrt(diag((rm.y'*rm.Kobs - rm.beta_hat'*V_star_inv*rm.beta_hat))'/(rm.n+rm.d-2));
%Kobs1 = L'\(L\(obs-H*beta_hat));
rm.Kobs1 = rm.Kinv*(rm.y-rm.H*rm.beta_hat);


rm.predict = @predict;
rm.predictGradient = @predict;
rm.setGrid = @setGrid;
rm.changeObs = @changeObs;
rm.predictGrid = @predictGrid;



rm.hasgrid = false;
rm.gridsize=0;
rm.grid = [];
rm.grid_k = [];
rm.grid_h = [];




    function rmout = setGrid(grid,rm) 
        rmout  = rm;
        rmout.gridsize = size(grid,1);
        rmout.hasgrid = true;
        rmout.grid = grid;
        rmout.grid_k  = zeros(rm.N,rmout.gridsize);
        rmout.grid_h  = zeros(rmout.gridsize,rm.q);
        for ii=1:rmout.gridsize
            for jj=1:rm.n
               JJ = 1 + (jj-1)*rm.step;
               rmout.grid_k(JJ:(JJ+rm.step-1),ii) = rm.kernel.evalPred(grid(ii,:),rm.input(jj,:),rm.kernel);
            end
            rmout.grid_h(ii,:) = rm.basisFunc(grid(ii,:));
        end
    end

    function rmout = changeObs(obs,obs1,rm)
        rmout = rm;
        if rmout.kernel.derivative
            rmout.y = zeros(rmout.N,1);
            for i=1:rmout.n
                I = 1 + (i-1)*rmout.step;
                rmout.y(I:(I+rmout.step-1)) = [obs(i);obs1(1,:,i)'];
            end
        else
            rmout.y = obs';
        end
        %Kobs = L'\(L\obs);
        rmout.Kobs = rmout.Kinv*rmout.y;
        %beta_hat = Lstar'\(Lstar\(H'*Kobs));
        rmout.beta_hat = rmout.V_star*(rmout.H'*rmout.Kobs);
        sigma2_hat = (rmout.y'*rmout.Kobs - rmout.beta_hat'*V_star_inv*rmout.beta_hat)/(rmout.n+rmout.d-2);
        if sigma2_hat < 0
            sigma2_hat = 1e-16;
        end
        rmout.sigma_hat = sqrt(sigma2_hat);
        %rmout.sigma_hat = sqrt(diag((rmout.y'*rmout.Kobs - rmout.beta_hat'*V_star_inv*rmout.beta_hat))'/(rmout.n+rmout.d-2));
        %Kobs1 = L'\(L\(obs-H*beta_hat));
        rmout.Kobs1 = rmout.Kinv*(rmout.y-rmout.H*rmout.beta_hat);
    end
        
    function [p,l,u] = predictGrid(j,rm) 
        if (~rm.hasgrid || j > rm.gridsize)
            p=0; u=0; l=0;
            return;
        end
        k = rm.grid_k(:,j);
        hx = rm.grid_h(j,:);
        s = 1;
        m_star = hx*rm.beta_hat + k'*rm.Kobs1;
        %Kinvk = L'\(L\k);
        Kinvk = rm.Kinv*k;

        hx1 = hx - k'*rm.KinvH;
        c_star = s - k'*Kinvk + hx1*rm.V_star*hx1';

       if c_star < 0
           c_star = abs(c_star);
       end

       pm = rm.conf*rm.sigma_hat*sqrt(c_star);
       u = m_star + pm;
       l = m_star - pm;
       p = m_star; 
    end
        
        


    function [p,l,u] = predict(x,rm) 
        k = zeros(rm.N,1);
        for jj=1:rm.n
           JJ = 1 + (jj-1)*rm.step;
           k(JJ:(JJ+rm.step-1)) = rm.kernel.evalPred(x,rm.input(jj,:),rm.kernel);
        end
        s = 1; 
        hx = rm.basisFunc(x);

        m_star = hx*rm.beta_hat + k'*rm.Kobs1;
        %Kinvk = L'\(L\k);
        Kinvk = rm.Kinv*k;

        hx1 = hx - k'*rm.KinvH;
        c_star = s - k'*Kinvk + hx1*rm.V_star*hx1';

       if c_star < 0
           c_star = abs(c_star);
       end

       pm = rm.conf*rm.sigma_hat*sqrt(c_star);
       u = m_star + pm;
       l = m_star - pm;
       p = m_star; 
    end


end

