function [ model ] = SIR_model( population_size )
%SIR_model generates the SIR model


model.var_names = {'S', 'I'}; %R has been reduced
model.par_names = {'ki','ke','kr','ks','N'};
model.dimension = 2;
model.parameter_dimension = 5;
model.observables_dimension = 3;

N = population_size;
model.var_init = [N;0];
model.par = [0.1/N; %ki - internal infection rate
             0.01;  %ke - external infection rate
             0.05;  %kr - recovery rate
             0.01;  %ks - immunity loss rate
             N];    %N - population_size

model.init_var_range = [floor(9*N/10),N;
                   0,N-floor(9*N/10)];         
model.par_range = [0.05/N,0.15/N;
                   0.005,0.015;
                   0.01,0.1;
                   0.005,0.015;
                   90,110];
model.observable_range = [0,N;
                          0,N;
                          0,N];               
         
model.explore_vars = logical([0;0]);
model.explore_pars = logical([1;0;0;0;0]);


model.vectorField = @vectorField;
model.jacobian = @jacobian;
model.sensitivityMatrix = @sensitivityMatrix;
model.observables = @observables;
model.sensitivity = @sensitivity;

model.odefunc = @ode;
model.reshapeInput = @reshapeInput;

model.split = @split;
model.merge = @merge;
model.enlargeInitVars = @enlargeInitVars;



    function y = vectorField(x,p) 
        rate = [p(1)*x(1)*x(2);             %internal infection
                p(2)*x(1);                  %external infection
                p(3)*x(2);                  %recovery
                p(4)*(p(5)-x(1)-x(2))];     %immunity loss  
       
        stoich = [ -1 -1  0  1;
                    1  1 -1  0];        
        y = stoich*rate;
    end


    function J = jacobian(x,p)
        stoich = [ -1 -1  0  1;
                    1  1 -1  0];
        %Jrate = d rate / d x;
        Jrate =  [p(1)*x(2), p(1)*x(1);
                       p(2),         0;
                          0,      p(3);
                      -p(4),     -p(4)];
        J = stoich*Jrate;
    end

    %this computes dF / dp, the derivative of the VF w.r.t. p
    function Jp = sensitivityMatrix(x,p)
        stoich = [ -1 -1  0  1;
                    1  1 -1  0];
        Jp_rate = [x(1)*x(2),0,0,0,0;             
                   0,x(1),0,0,0;                  
                   0,0,x(2),0,0;                  
                   0,0,0,p(5)-x(1)-x(2),p(4)];
        Jp = stoich*Jp_rate;        
    end


    function dvdt = ode(v,p,t,model) 
       [x,Sx,Sp] = model.split(v,model);
       dxdt = model.vectorField(x,p);
       J = model.jacobian(x,p);
       Jp = model.sensitivityMatrix(x,p);
       dSpdt = J*Sp + Jp;
       dSxdt = J*Sx;
       dvdt = model.merge(dxdt,dSxdt,dSpdt);
        
        
    end


    function [x,Sx,Sp] = split(v,model)
        n = model.dimension;
        m = model.parameter_dimension;
        x = v(1:n);
        Sx = reshape(v((n+1):(n+n*n)),n,n);
        Sp = reshape(v((n+n*n+1):end),n,m);
    end

    function v = merge(x,Sx,Sp)
       v = [x;Sx(:);Sp(:)];        
    end


    function o = observables(v,model)
        [x,~,~] = model.split(v,model);
        %here I observe  the individual variables and their sum
        o = [x(1);
             x(2);
             x(1)+x(2)];
    end
        
    %s is a matrix, each ro containing the derivative of observables w.r.t 
    %each variable and parameter.
    function S = sensitivity(v,p,model)
        [x,Sx,Sp] = model.split(v,model);
        %derivative of the observables w.r.t. x Jo(i,j) = do(i)/dx(j)
        Jo = [1,0;
              0,1;
              1,1];
        %returns sensitivity only w.r.t. explored parameters  
        ox = Jo*Sx(:,model.explore_vars);  
        op = Jo*Sp(:,model.explore_pars);
        o1 = Jo*vectorField(x,p);
        S = [ox,op,o1];
    end

 
    function [x0,p] = reshapeInput(input,model)
        n = model.dimension;
        m = model.parameter_dimension;
        n1 = sum(model.explore_vars);      
     
        x0 = zeros(n,1);
        x0(model.explore_vars) = input(1:n1);
        x0(~model.explore_vars) = model.var_init(~model.explore_vars);
        p = zeros(m,1);
        p(model.explore_pars) = input(n1+1:end);
        p(~model.explore_pars) = model.par(~model.explore_pars);
    end
                    
    function v0 = enlargeInitVars(x0,model)
        n = model.dimension;
        m = model.parameter_dimension;
        Sx = eye(n,n);
        Sp = zeros(n,m);
        v0 = model.merge(x0,Sx,Sp);
    end


end

