function [ model ] = getModel( modelSym )
%SIR_model generates a model from a symbolic description of the model
%itself

for i=1:length(modelSym.var)
    model.var_names{i} = char(modelSym.var(i));
end
for i=1:length(modelSym.par)
    model.par_names{i} = char(modelSym.par(i));
end

model.dimension = length(modelSym.var);
model.parameter_dimension = length(modelSym.par);
model.observables_dimension = length(modelSym.obs);

model.var_init = modelSym.var_init;
model.par = modelSym.par_values;

model.init_var_range = modelSym.init_var_range;         
model.par_range = modelSym.par_range;
model.observable_range = modelSym.obs_range;               
model.explore_vars = logical(modelSym.explore_vars);
model.explore_pars = logical(modelSym.explore_pars);

xs = 'x(1)';
for i=2:model.dimension
    xs = [xs,',x(',int2str(i),')'];
end
xp = xs;
for i=1:model.parameter_dimension
    xp = [xp,',p(',int2str(i),')'];
end

F = modelSym.stoichiometry*modelSym.rate;
eF = matlabFunction(F,'vars',[modelSym.var,modelSym.par]);
eval(['model.vectorField = @(x,p)eF(',xp,');']);

J = jacobian(F,modelSym.var);
eJ = matlabFunction(J,'vars',[modelSym.var,modelSym.par]);
eval(['model.jacobian = @(x,p)eJ(',xp,');']);

JP = jacobian(F,modelSym.par);
eJp = matlabFunction(JP,'vars',[modelSym.var,modelSym.par]);
eval(['model.sensitivityMatrix = @(x,p)eJp(',xp,');']);

eO = matlabFunction(modelSym.obs,'vars',modelSym.var);
eval(['model.obsF = @(x)eO(',xs,');']);

JO = jacobian(modelSym.obs,modelSym.var);
eJO = matlabFunction(JO,'vars',modelSym.var);
eval(['model.Jobs = @(x)eJO(',xs,');']);

oP = JO*F;
eoP = matlabFunction(oP,'vars',[modelSym.var,modelSym.par]);
eval(['model.obsPrime = @(x,p)eoP(',xp,');']);

model.observables = @observables;
model.sensitivity = @sensitivity;

model.odefunc = @ode;
model.reshapeInput = @reshapeInput;

model.split = @split;
model.merge = @merge;
model.enlargeInitVars = @enlargeInitVars;




    function o = observables(v,model)
        [x,~,~] = model.split(v,model);
        %here I observe  the individual variables and their sum
        o = model.obsF(x);
    end


    function dvdt = ode(v,p,t,model) 
       [x,Sx,Sp] = model.split(v,model);
       dxdt = model.vectorField(x,p);
       J = model.jacobian(x,p);
       Jp = model.sensitivityMatrix(x,p);
       dSpdt = J*Sp + Jp;
       dSxdt = J*Sx;
       dvdt = model.merge(dxdt,dSxdt,dSpdt);       
    end


    function [x,Sx,Sp] = split(v,model)
        n = model.dimension;
        m = model.parameter_dimension;
        x = v(1:n);
        Sx = reshape(v((n+1):(n+n*n)),n,n);
        Sp = reshape(v((n+n*n+1):end),n,m);
    end

    function v = merge(x,Sx,Sp)
       v = [x;Sx(:);Sp(:)];        
    end

        
    %s is a matrix, each ro containing the derivative of observables w.r.t 
    %each variable and parameter.
    function S = sensitivity(v,p,model)
        [x,Sx,Sp] = model.split(v,model);
        %derivative of the observables w.r.t. x Jo(i,j) = do(i)/dx(j)
        Jo = model.Jobs(x);
        %returns sensitivity only w.r.t. explored parameters  
        ox = Jo*Sx(:,logical(model.explore_vars));  
        op = Jo*Sp(:,logical(model.explore_pars));
        o1 = model.obsPrime(x,p);
        S = [ox,op,o1];
    end

 
    function [x0,p] = reshapeInput(input,model)
        n = model.dimension;
        m = model.parameter_dimension;
        n1 = sum(model.explore_vars);      
     
        x0 = zeros(n,1);
        x0(logical(model.explore_vars)) = input(1:n1);
        x0(~logical(model.explore_vars)) = model.var_init(~model.explore_vars);
        p = zeros(m,1);
        p(logical(model.explore_pars)) = input(n1+1:end);
        p(~logical(model.explore_pars)) = model.par(~model.explore_pars);
    end
                    
    function v0 = enlargeInitVars(x0,model)
        n = model.dimension;
        m = model.parameter_dimension;
        Sx = eye(n,n);
        Sp = zeros(n,m);
        v0 = model.merge(x0,Sx,Sp);
    end


    


end

