function [selected,maxvalue,maxstddev,maxucb,extra,diagnostic] =...
         GPUCB_Optimise(func,domain,kernel,options)
%
% GPUCB_Optimise GP_UCB optimisation with adaptive grid
% 
% Uses the GP_UCB algorithm to maximise likelihood of parameters. 
% Incorporates an adaptive selection of grid points to speed it up. 
% (guaranteeing convergence in expectation?)
% 
% Input arguments
%   func        this is a function handle to the function to optimise. 
%               It takes one parameters:
%               * newInput  - a vector containing the point of the input 
%                             space for which the function has to be 
%                             evaluated 
%               It returns two values: the estimated function value and an 
%               estimate of the observation noise.
%              
%   inputSpace  a structure containing information about the input
%               space. It specifically contains the following fields:
%               * dimension - the dimension of the input space
%               * trueRange - a dimension x 2 matrix, containing for each
%                             coordinate the lower and upper bound on 
%                             its range
%               * standardise  - a boolean, true if standardisation has to
%                                be performed
%               * log       - a dimension (0/1) column vector, equal to 1
%                             in position j if we need to take the 
%                             log scale for coordinate j
%               * range     - a dimension x 2 matrix, containing for each
%                             parameter the lower and upper bound on 
%                             the standardised range
%               * stdF      - the standardisation function, taking as input
%                             a matrix of parameters (which are in the column 
%                             vectors), and a params structure
%               * invStdF   - the inverse standardisation function, taking 
%                             a matrix of parameters (which are in the column 
%                             vectors), and a params structure
%               * names     - (optional) a cell array with parameter names.
%   options     a structure containing the options for the optmization
%               algorithm. It contains the following fields:
%               * gridSampling  -  a boolean (0/1) value. If true, the grid
%                   of new points is sampled adaptively at each iteration of
%                   the algorithm. Use in high dimensional spaces.
%               * gridSampleNumber  -  an integer, containing the number of 
%                   grid points to be sampled.
%               * maxGridResamples - maximum number of times a grid is
%                   resampled without improvements
%               * minNewPoints - minimum number of points added to inputs
%                   by the algorithm.
%               * initialObs - number of initial observations (if on a
%                   fixed grid, these are per parameter)
%               * randomInitialPoints - a boolean 0/1 value. If true, the
%                   initial observations are sampled from a randon grid.
%               * sampleInitialPointsFromPrior - samples initial points
%                   from the prior distribution
%               * probSamplingFromPrior - the probability of sampling from
%                   prior, if sampleInitialPointsFromPrior is true.
%               * extraGrid  -  an integer, containing the number of
%                   extra grid points to sample around the chosen point of
%                   the grid, to increase exploitation of the search.
%               * dispersion - the standard deviation of the
%                   normal distribution used to sample the extra grid
%                   points around the chosen one.
%               * extraPointsForHessian - number of extra evaluations of the 
%                   likelihood in order to improve the estimate of the variance
%                   via the hessian
%               * extraPointsDispersion - Gaussian dispersion for the extra
%                   points sampled for the Hessian
%               * smcSamples  -  the number of trajectories to be sampled
%                   by the statistical model checker. This is to be used by
%                   the likelihood function likeFunc.
%               * kernel  -  a structure containing information about the
%                   kernel defining the GP. To be used by
%                   propCheckKernCompute. Refer to its documentation for
%                   the fields of kernel
%               * debug  -  a boolean (0/1) value, if true debug
%                   information is visualised during the optimisation.
%               * verbose - how much info is shown during debug
%               * debugPlot  -  shows plots of the current GP function. 
%                   Works only for 1D parameter spaces. 
%               * betaType - the type of beta_t parameter of the
%                   algortithm. Currently we have: 'const' (constant
%                   value), 'loginc' (a value incrementing logarithmically 
%                   with the number of iterations), and 'repinflate' (a
%                   constant value inflated if repeating the grid sampling
%                   while looking for convergence).
%               * betaConst - the costant value of beta for 'const' and
%                   'repinflate'
%               * betaDelta - delta parameter for 'loginc'. Has to be in (0,1)
%               * betaRepMult - multiplicative factor for repetitions in
%               'repinflate'
%               * betaFunc - the function computing beta, takes 4 inputs:
%                   number of iterations, number of repetitions, params,
%                   options
%               * usePrior - a 0/1 value. If true, a prior is used in
%                   computing the loglikelihood function.
%               * maxAddedPointsNoImprovement - controls how many new
%                   points can be added to the observations without
%                   substantially improving the predicted value of the
%                   decision function in the new point. Allows to terminate
%                   the algorithm if the maximum is in a large plateau
%               * improvementFactor - the factor > 1 governing how large 
%                   an improvement has to be to be considered substantial.
%                   The improvement is relative to the previous maximum.
%
% Output arguments
%   selected    the point of the input space maximising the likelihood
%   maxvalue    the predicted value at the maximum 
%   maxstddev   the standard deviation at the max value.
%   diagnostic  a structure cointaing auxiliary information on the
%               execution of the algorithm. It contains the following
%               fields: 
%               * likeEval  -  the number of times the likelihood has been
%                              evaluated it the main loop of the algorithm
%               * extraLikeEval - the extra likelihood evaluations done by
%                               the algorithm for computing the Hessian
%               * totalTime  -  the total execution time of the optimisation
%               * likeEvalTime - the time for the evaluation of the
%                                likelihood function
%   extra       a structure containing extra information:
%               * inputValues
%               * obsValues
%               * gridPoints
%               * regression model



tID = tic;
diagnostic.funcEval = 0;
diagnostic.funcEvalTime = 0;
% generate the initial observation points, either randomly or on a fixed grid
if options.randomInitialPoints 
    if options.orthogonalSamplingInitialPoints
        inputVals = GPUCB_gridOrthogonalSample(domain,...
            domain.dimension,options.othogonalSampleBinsInit,...
            options.orthogonalSamplePointsPerBinInit);
    else
        inputVals = GPUCB_gridSample(options.initialObs,domain);
    end
else
    inputVals = GPUCB_fixedGrid(options.initialObs,domain);
end

% compute initial observations
numInpts = size(inputVals,2);
obsData = zeros(1,numInpts);
noise = zeros(1,numInpts);
for i =1:numInpts
    if (domain.standardise)
        p = domain.invStdF(inputVals(:,i),domain);
    else
        p = inputVals(:,i);
    end
    ttid = tic;
    [obsData(i),noise(i)] = func(p);
    tt = toc(ttid);
    
    if options.debug
        disp(['Initial likelihood evaluation ', int2str(i),' took time: ',num2str(tt)]);
    end
    diagnostic.funcEval = diagnostic.funcEval + 1;
    diagnostic.funcEvalTime = diagnostic.funcEvalTime + tt;
end

if options.debugPlot && domain.dimension == 1
   figure, plot(inputVals,obsData,'.-'); 
   disp('Plotted initial grid with computed likelihood');
   disp('Press enter to continue');
   pause
end

%adapting kernel values
if options.hyperparameterOptimisation
    [a,l] =  GPUCB_optimiseHyperparameters(inputVals,obsData,noise,...
                kernel,options.hyperparameterOptTrials);
    kernel.amplitude = a;
    kernel.lengthScale = l;
    
    if options.debug
        disp(['Optmised hyperparameters: amplitude = ', num2str(a),...
            '; lenghtscale = ',num2str(l)]);
    end
end



if options.gridSampling
    if options.gridOrthogonalSampling
        gridVals = GPUCB_gridOrthogonalSample(domain,...
            domain.dimension,options.othogonalSampleBins,...
            options.orthogonalSamplePointsPerBin);
    else
      %this samples uniformly from the parameter range
        gridVals=GPUCB_gridSample(options.gridSampleNumber,domain);
    end
else
    gridVals=GPUCB_fixedGrid(options.gridSampleNumber,domain);
end

%adding extra grid points around one maximum of the current likelihood
if options.extraGrid
    %find the maximum in the obs data
    firstMax=inputVals(:,obsData==max(obsData));
    if size(firstMax,2)>1
        %just in case two points had exactly the same decision value
        %randomly samples one of the decision values
        z=randi(size(firstMax,2));
        firstMax=firstMax(:,z);            
    end
    % if we require extra grid points 
    %add a scattering of new grid points near the new point
    extraGridPts = firstMax*ones(1,options.extraGrid) + ...
        options.dispersion * randn(domain.dimension,options.extraGrid);
    extraGridInd = domain.belongs(extraGridPts,domain);
    newGridVals=[gridVals,extraGridPts(:,extraGridInd)];
    % at next step, forget the extra grid points we used unless one 
    % of them was the maximum         
else
    newGridVals=gridVals;
end

if options.debug
    disp(['Initial grid generated']);
end

resamplingCounter=0;
notConverged=1;
iterations = 0;
addedPoints = 0;
pointAddedLastIter = 0;
addedPointsNoImprovement = 0;


while notConverged
    iterations = iterations + 1;
    numInpts=size(inputVals,2);
    numGridPts=size(newGridVals,2);
%       %concatenates the inputs of observed values with the additional grid 
%       %points. The observed inputs are always first! 
     totalGridVals=[inputVals,newGridVals]; 
%       %make it zero mean to avoid excessive misfit with GPprior
%     obsData=obsDataOriginal-mean(obsDataOriginal);
 
       %computes the Gram matrix; this could be speeded up if needed as 
       %we do not need to recompute the whole Gram matrix every time.
       rm = GPregressionMean(inputVals',obsData,noise,kernel);
       [GPmean,GPstd] = rm.predictSet(newGridVals',rm); 
       [GPmeanObs,GPstdObs] = rm.predictSet(inputVals',rm);

    if options.debugPlot && domain.dimension == 1
        %shows diagnostic plots if data is 1d
        figure, plot(inputVals,obsData,'r+');
        hold on
        [sortedGrid,sortInd]=sort(totalGridVals,'ascend');%sorts grid values for plotting
        mm = [GPmeanObs; GPmean];
        vv = [GPstdObs;GPstd];
        bb = options.betaFunc(iterations,resamplingCounter,domain,options);
        plot(sortedGrid,mm(sortInd))
        hold on
        plot(sortedGrid,mm(sortInd)+bb*(vv(sortInd)),'k.-')
        plot(sortedGrid,mm(sortInd)-bb*(vv(sortInd)),'k.-')
        disp('Plotted estimated likelihood and uncertainty regions');
        disp('Press enter to continue');
        pause
    end
    
    %computes the decision function for grid and input points, which 
    %incorporates information about the mean (best points predicted so far)
    %and the variance (unknown regions).
    beta = options.betaFunc(iterations,resamplingCounter,domain,options);
    decisionObs = GPmeanObs + beta*GPstdObs;
    decision = GPmean + beta*GPstd;
    
    if options.debug && options.verbose
        disp('Grid points and estimated maximum.')
        if (domain.standardise)
            xx=domain.invStdF(totalGridVals,domain);
        else
            xx=totalGridVals;
        end
        disp([xx', [obsData'; decision] ]);
    end
    %check if we have increased the max value reasonably
    if pointAddedLastIter 
        if decisionObs(end) <= options.improvementFactor*max(decisionObs(1:end-1))
            addedPointsNoImprovement = addedPointsNoImprovement + 1;
        else
            addedPointsNoImprovement = 0;
        end
    end
    %find the maximum value of obsData and of the decision function
    maxObs = max(decisionObs);
    maxDec = max(decision);
    
    
    if addedPointsNoImprovement >= options.maxAddedPointsNoImprovement
        %the solution improved too few for too many times, we stop.
        notConverged=0;
        %find the index of the maximum in obsData
        maxIndex = find(decisionObs==max(decisionObs));
        %correct a strange effect: sometimes there are two identical
        %points, both maxima.
        maxIndex = maxIndex(1);
        %if fixed grid we are finished
       
        if options.debug
            disp('No new maximum found. Terminating.');
            disp('Maximum found at:')
            if (domain.standardise)
                xx=domain.invStdF(totalGridVals(:,maxIndex),domain);
            else
                xx=totalGridVals(:,maxIndex);
            end
            disp(xx);
        end
        
    elseif maxDec >= maxObs || addedPoints < options.minNewPoints
        pointAddedLastIter = 1;
        %found a new potential maximum, identifying its index
        %indexNewInput= numInpts + find(decision==max(decision));
        indexNewInput= numInpts + find(decision==max(decision));
        resamplingCounter=0;
        addedPoints = addedPoints + 1;
        %finds the grid point which has the maximum mean+2stDev
        newInput=totalGridVals(:,indexNewInput);
        if size(newInput,2)>1
            %just in case two points had exactly the same decision value
            %randomly samples one of the decision values
            z=randi(size(newInput,2));
            newInput=newInput(:,z);            
        end
        %here we can run a local optimisation, to find a proper maximum
        %nearby the selected point, and add it as the new input.
        if (options.localOptimisation)
            [newInput,~] = GPUCB_localOpt(rm,beta,newInput,domain);
        end
        
        inputVals=[inputVals,newInput];
        %computes the likelihood via SMC and adds it to the observed data
        if domain.standardise
            newP = domain.invStdF(newInput,domain);
        else
            newP = newInput;
        end
        ttid = tic;
        [newData,newNoise] = func(newP);
        time = toc(ttid);
        %increase function evaluation counter and time tracker
        diagnostic.funcEval = diagnostic.funcEval + 1;
        diagnostic.funcEvalTime = diagnostic.funcEvalTime + time;
        
        obsData=[obsData,newData];
        noise = [noise,newNoise];
        
    
        if options.gridSampling
            if options.gridOrthogonalSampling
                 gridVals = GPUCB_gridOrthogonalSample(domain,...
                    domain.dimension,options.othogonalSampleBins,...
                    options.orthogonalSamplePointsPerBin);
            else
              %this samples uniformly from the parameter range
                gridVals=GPUCB_gridSample(options.gridSampleNumber,domain);
            end
            %resamples the grid                   
        else
            if indexNewInput-numInpts <= size(gridVals,2)
                gridVals = [gridVals(:,1:indexNewInput-numInpts-1),gridVals(:,indexNewInput-numInpts+1:end)];
                %else we chose a point from the extra grid points!
            end
                %Point is removed from the grid, if it was in the grid!               
        end
        
        if options.extraGrid
            % if we require extra grid points 
            %add a scattering of new grid points near the new point
            extraGridPts = newInput*ones(1,options.extraGrid) + ...
                options.dispersion * randn(domain.dimension,options.extraGrid);                
            extraGridInd = domain.belongs(extraGridPts,domain);
            newGridVals=[gridVals,extraGridPts(:,extraGridInd)];
            % at next step, forget the extra grid points we used unless one 
            % of them was the maximum         
        else
            newGridVals=gridVals;
        end
                
        
        if options.debug
            disp(['GP_UCB iteration ',int2str(diagnostic.funcEval)]);
            disp('Found new candidate maximum:');
            if (domain.standardise)
                xx=domain.invStdF(newInput,domain);
            else
                xx=newInput;
            end
            disp(xx);
        end
    else
        pointAddedLastIter = 0;
        %the maximum decision point is among the observed ones; 
        %converge if fixed grid or resample the grid a few times 
        %and then declare it converged
        if options.gridSampling
            resamplingCounter=resamplingCounter+1;
            if resamplingCounter>options.maxGridResamples
                %find the index of the maximum in obsData
                maxIndex = find(decisionObs==max(decisionObs));
                %correct a strange effect: sometimes there are two identical
                %points, both maxima.
                maxIndex = maxIndex(1);
                
                if options.debug
                    disp('No new maximum found. Terminating.');
                    disp('Maximum found at:')
                    if (domain.standardise)
                        xx=domain.invStdF(totalGridVals(:,maxIndex),domain);
                    else
                        xx=totalGridVals(:,maxIndex);
                    end
                    disp(xx);
                end
                
                notConverged=0; %finished
            else
                if options.gridOrthogonalSampling
                    gridVals = GPUCB_gridOrthogonalSample(domain,...
                        domain.dimension,options.othogonalSampleBins,...
                        options.orthogonalSamplePointsPerBin);
                else
                  %this samples uniformly from the parameter range
                    gridVals=GPUCB_gridSample(options.gridSampleNumber,domain);
                end
               %resample grid while the counter is small enough
                newGridVals = gridVals;
                
                if options.debug
                    disp('No new maximum found. Sampling a new grid.');
                end
            end
        else
            notConverged=0;
            %find the index of the maximum in obsData
            maxIndex = find(decisionObs==max(decisionObs));
            %correct a strange effect: sometimes there are two identical
            %points, both maxima.
            maxIndex = maxIndex(1);
            
            %if fixed grid we are finished
            if options.debug
                disp('No new maximum found. Terminating.');
                disp('Maximum found at:')
                if (domain.standardise)
                    xx=domain.invStdF(totalGridVals(:,maxIndex),domain);
                else
                    xx=totalGridVals(:,maxIndex);
                end
                disp(xx);
            end
        end
    end
end
%return maximum, maxvalue, and so on
if (domain.standardise)
    selected=domain.invStdF(totalGridVals(:,maxIndex),domain);
else
    selected=totalGridVals(:,maxIndex);
end
[maxvalue,maxstddev] = rm.predict(totalGridVals(:,maxIndex),rm);
maxucb = maxvalue +  rm.conf * maxstddev;

%store auxiliary information
if (domain.standardise)
    extra.inputValues=domain.invStdF(inputVals,domain);
else
    extra.inputValues=inputVals;
end 
extra.obsValues = obsData;
extra.noise = noise;
if (domain.standardise)
    extra.gridPoints=domain.invStdF(newGridVals,domain);
else
    extra.gridPoints=newGridVals;
end
extra.estimatedFunction = decision;
extra.GPmean = [GPmeanObs; GPmean];
extra.GPstd =[GPstdObs;GPstd];

diagnostic.totalTime = toc(tID);
%print informations at the end of the algorithm
disp('*******************************************************************************************');
disp('Optimisation finished!');
disp('Found maximum of function at point:');
disp(selected);
disp(['Predicted function value at maximum is ',num2str(maxvalue)]);
disp(['Predicted function stdev at maximum is ',num2str(maxstddev)]);
disp(['Predicted function UCB at maximum is ',num2str(maxucb)]);
disp(['Function evaluations: ', num2str(diagnostic.funcEval)]);
disp(['Time taken by the optimisation: ', num2str(diagnostic.totalTime), ' seconds']);
disp(['Time taken to evaluate the function: ', num2str(diagnostic.funcEvalTime), ' seconds']);
disp('*******************************************************************************************');    



    
    

    
    